"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts = require("typescript");
const ymir_1 = require("@fimbul/ymir");
const inversify_1 = require("inversify");
const core_module_1 = require("../src/di/core.module");
const default_module_1 = require("../src/di/default.module");
const configuration_manager_1 = require("../src/services/configuration-manager");
const linter_1 = require("../src/linter");
const utils_1 = require("../src/utils");
const cached_file_system_1 = require("../src/services/cached-file-system");
const resolve = require("resolve");
const path = require("path");
const yaml = require("js-yaml");
const argparse_1 = require("../src/argparse");
const normalize_glob_1 = require("normalize-glob");
const minimatch_1 = require("minimatch");
exports.version = '1';
class LanguageServiceInterceptor {
    constructor(config, 
    // tslint:disable:no-submodule-imports
    project, serverHost, 
    // tslint:enable:no-submodule-imports
    languageService, require, log) {
        this.config = config;
        this.project = project;
        this.serverHost = serverHost;
        this.languageService = languageService;
        this.require = require;
        this.log = log;
        this.findingsForFile = new WeakMap();
    }
    updateConfig(config) {
        this.config = config;
    }
    getSemanticDiagnostics(diagnostics, fileName) {
        const program = this.languageService.getProgram();
        const file = program.getSourceFile(fileName);
        if (file === undefined) {
            this.log(`file ${fileName} is not included in the Program`);
        }
        else {
            this.log(`started linting ${fileName}`);
            const findings = this.getFindings(file, program);
            this.findingsForFile.set(file, findings);
            diagnostics = diagnostics.concat(findings.map((finding) => ({
                file,
                category: finding.severity === 'error'
                    ? this.config.displayErrorsAsWarnings
                        ? ts.DiagnosticCategory.Warning
                        : ts.DiagnosticCategory.Error
                    : finding.severity === 'warning'
                        ? ts.DiagnosticCategory.Warning
                        : ts.DiagnosticCategory.Suggestion,
                code: finding.ruleName,
                source: 'wotan',
                messageText: finding.message,
                start: finding.start.position,
                length: finding.end.position - finding.start.position,
            })));
            this.log(`finished linting ${fileName} with ${findings.length} findings`);
        }
        return diagnostics;
    }
    getFindings(file, program) {
        let globalConfigDir = this.project.getCurrentDirectory();
        let globalOptions;
        while (true) {
            const scriptSnapshot = this.project.getScriptSnapshot(globalConfigDir + '/.fimbullinter.yaml');
            if (scriptSnapshot !== undefined) {
                this.log(`Using '${globalConfigDir}/.fimbullinter.yaml' for global options.`);
                globalOptions = yaml.safeLoad(scriptSnapshot.getText(0, scriptSnapshot.getLength())) || {};
                break;
            }
            const parentDir = path.dirname(globalConfigDir);
            if (parentDir === globalConfigDir) {
                this.log("Cannot find '.fimbullinter.yaml'.");
                globalOptions = {};
                break;
            }
            globalConfigDir = parentDir;
        }
        const globalConfig = argparse_1.parseGlobalOptions(globalOptions);
        if (!isIncluded(file.fileName, globalConfigDir, globalConfig))
            return [];
        const container = new inversify_1.Container({ defaultScope: inversify_1.BindingScopeEnum.Singleton });
        for (const module of globalConfig.modules)
            container.load(this.loadPluginModule(module, globalConfigDir, globalOptions));
        container.bind(ymir_1.FileSystem).toConstantValue(new ProjectFileSystem(this.project));
        container.bind(ymir_1.DirectoryService).toConstantValue({
            getCurrentDirectory: () => this.project.getCurrentDirectory(),
        });
        container.bind(ymir_1.Resolver).toDynamicValue((context) => {
            const fs = context.container.get(cached_file_system_1.CachedFileSystem);
            return {
                getDefaultExtensions() {
                    return ['.js'];
                },
                resolve(id, basedir, extensions = ['.js'], paths) {
                    return resolve.sync(id, {
                        basedir,
                        extensions,
                        paths,
                        isFile: (f) => fs.isFile(f),
                        readFileSync: (f) => fs.readFile(f),
                    });
                },
                require: this.require,
            };
        });
        const warnings = [];
        container.bind(ymir_1.MessageHandler).toConstantValue({
            log: this.log,
            warn: (message) => {
                if (utils_1.addUnique(warnings, message))
                    this.log(message);
            },
            error(e) {
                this.log(e.message);
            },
        });
        container.load(core_module_1.createCoreModule(globalOptions), default_module_1.createDefaultModule());
        const fileFilter = container.get(ymir_1.FileFilterFactory).create({ program, host: this.project });
        if (!fileFilter.filter(file))
            return [];
        const configManager = container.get(configuration_manager_1.ConfigurationManager);
        const config = globalConfig.config === undefined
            ? configManager.find(file.fileName)
            : configManager.loadLocalOrResolved(globalConfig.config, globalConfigDir);
        const effectiveConfig = config && configManager.reduce(config, file.fileName);
        if (effectiveConfig === undefined)
            return [];
        const linter = container.get(linter_1.Linter);
        return linter.lintFile(file, effectiveConfig, program, {
            reportUselessDirectives: globalConfig.reportUselessDirectives
                ? globalConfig.reportUselessDirectives === true
                    ? 'error'
                    : globalConfig.reportUselessDirectives
                : undefined,
        });
    }
    loadPluginModule(moduleName, basedir, options) {
        moduleName = resolve.sync(moduleName, {
            basedir,
            extensions: ['.js'],
            isFile: (f) => this.project.fileExists(f),
            readFileSync: (f) => this.project.readFile(f),
        });
        const m = this.require(moduleName);
        if (!m || typeof m.createModule !== 'function')
            throw new Error(`Module '${moduleName}' does not export a function 'createModule'.`);
        return m.createModule(options);
    }
    getSupportedCodeFixes(fixes) {
        return fixes;
    }
    dispose() {
        // TODO clean up after ourselves
        return this.languageService.dispose();
    }
}
exports.LanguageServiceInterceptor = LanguageServiceInterceptor;
function isIncluded(fileName, basedir, options) {
    outer: if (options.files.length !== 0) {
        for (const include of options.files)
            for (const normalized of normalize_glob_1.normalizeGlob(include, basedir))
                if (new minimatch_1.Minimatch(normalized).match(fileName))
                    break outer;
        return false;
    }
    for (const exclude of options.exclude)
        for (const normalized of normalize_glob_1.normalizeGlob(exclude, basedir))
            if (new minimatch_1.Minimatch(normalized, { dot: true }).match(fileName))
                return false;
    return true;
}
class ProjectFileSystem {
    constructor(host) {
        this.host = host;
        this.realpath = this.host.realpath && ((f) => this.host.realpath(f));
    }
    createDirectory() {
        throw new Error('should not be called.');
    }
    deleteFile() {
        throw new Error('should not be called.');
    }
    writeFile() {
        throw new Error('should not be called.');
    }
    normalizePath(f) {
        f = f.replace(/\\/g, '/');
        return this.host.useCaseSensitiveFileNames() ? f : f.toLowerCase();
    }
    readFile(f) {
        const result = this.host.readFile(f);
        if (result === undefined)
            throw new Error('ENOENT');
        return result;
    }
    readDirectory(dir) {
        return this.host.readDirectory(dir, undefined, undefined, ['*']);
    }
    stat(f) {
        const isFile = this.host.fileExists(f)
            ? true
            : this.host.directoryExists(f)
                ? false
                : undefined;
        return {
            isDirectory() {
                return isFile === false;
            },
            isFile() {
                return isFile === true;
            },
        };
    }
}
//# sourceMappingURL=index.js.map